@@ -1,5 +1,5 @@
 # Copyright 2019 Katteli Inc.
-# TestFlows Test Framework (http://testflows.com)
+# TestFlows.com Open-Source Software Testing Framework (http://testflows.com)
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -17,6 +17,7 @@ from testflows._core.cli.arg.common import HelpFormatter
 from testflows._core.cli.arg.handlers.handler import Handler as HandlerBase
 from testflows._core.cli.arg.handlers.document.convert import Handler as convert_handler
 from testflows._core.cli.arg.handlers.document.toc import Handler as toc_handler
+from testflows._core.cli.arg.handlers.document.new.handler import Handler as new_handler
 
 class Handler(HandlerBase):
     @classmethod
@@ -30,3 +31,4 @@ class Handler(HandlerBase):
         document_commands.required = True
         convert_handler.add_command(document_commands)
         toc_handler.add_command(document_commands)
+        new_handler.add_command(document_commands)
