@@ -16,24 +16,23 @@ import os
 import sys
 import time
 import random
+import asyncio
 import inspect
 import builtins
-import operator
 import functools
-import tempfile
 import threading
-import textwrap
 import importlib
 
-from contextlib import ExitStack, contextmanager
+from contextlib import ExitStack
 from collections import namedtuple
 
 import testflows.settings as settings
+import testflows._core.contrib.yaml as yaml
 
 from .templog import filename as templog_filename
 from .exceptions import DummyTestException, ResultException, TestIteration, DescriptionError, TestRerunIndividually
 from .flags import Flags, SKIP, TE, FAIL_NOT_COUNTED, ERROR_NOT_COUNTED, NULL_NOT_COUNTED, MANDATORY, MANUAL, AUTO
-from .flags import PARALLEL, NO_PARALLEL
+from .flags import PARALLEL, NO_PARALLEL, ASYNC
 from .flags import XOK, XFAIL, XNULL, XERROR, XRESULT
 from .flags import EOK, EFAIL, EERROR, ESKIP, ERESULT
 from .flags import CFLAGS, PAUSE_BEFORE, PAUSE_AFTER
@@ -51,14 +50,18 @@ from .cli.arg.common import epilog as common_epilog
 from .cli.arg.exit import ExitWithError, ExitException
 from .cli.arg.type import key_value as key_value_type, repeat as repeat_type, tags_filter as tags_filter_type
 from .cli.arg.type import logfile as logfile_type, rsa_private_key_pem_file as rsa_private_key_pem_file_type
+from .cli.arg.type import file as file_type
 from .cli.arg.type import onoff as onoff_type, count as count_type
 from .cli.text import danger, warning
 from .exceptions import exception as get_exception
 from .filters import The, TheTags
 from .utils.sort import human as human_sort
 from .transform.log.pipeline import ResultsLogPipeline
-from .parallel import current, top, previous, _check_parallel_context
-from .parallel import Pool as ParallelPool, join as parallel_join
+from .parallel import current, top, previous, _check_parallel_context, join as parallel_join
+from .parallel import convert_result_to_concurrent_future
+from .parallel.executor.thread import ThreadPoolExecutor, GlobalThreadPoolExecutor
+from .parallel.executor.asyncio import AsyncPoolExecutor, GlobalAsyncPoolExecutor
+from .parallel.asyncio import is_running_in_event_loop
 
 try:
     import testflows.database as database_module
@@ -365,7 +368,9 @@ class TestBase(object):
             parallel_exception = None
             # join any left over parallel tests and save
             try:
-                parallel_join(*self.futures, cancel=(exc_value is not None), test=self)
+                # maybe we should raise an error if self.futures is not empty to force users
+                # to explicitly join their explicit and implicit parallel tests
+                parallel_join(*self.futures, cancel=(exc_value is not None), test=self, _awaited=False)
             except (Exception, KeyboardInterrupt) as exc:
                 parallel_exception = exc
 
@@ -373,9 +378,12 @@ class TestBase(object):
             if self.executor is not None:
                 self.executor.__exit__(None, None, None)
 
-            # close global parallel pool if present and opened
-            if top() is self and settings.global_parallel_pool is not None:
-                settings.global_parallel_pool.__exit__(None, None, None)
+            # close global pools if present and opened
+            if top() is self:
+                if settings.global_thread_pool is not None:
+                    settings.global_thread_pool.__exit__(None, None, None)
+                if settings.global_async_pool is not None:
+                    settings.global_async_pool.__exit__(None, None, None)
 
             # set parallel exception if exc_value is None
             if exc_value is None and parallel_exception is not None:
@@ -508,7 +516,7 @@ class TestBase(object):
         return self.io.message_io(name=name)
 
 epilog = """
-option values:
+argument values:
 
 pattern
   used to match test names using a unix-like file path pattern that supports wildcards
@@ -548,6 +556,10 @@ def cli_argparser(kwargs, argparser=None):
 
     parser = main_parser.add_argument_group("common arguments")
 
+    parser.add_argument("--config", "-c", dest ="_config", metavar="yml ...", action="append", default=[],
+                        help=("test run YAML configuration file. Can be specified more than once "
+                              "to apply multiple configuration files that are "
+                              "applied left to right"), type=file_type("r"))
     parser.add_argument("--name", dest="_name", metavar="name",
                         help="test run name", type=str, required=False)
     parser.add_argument("--tag", dest="_tags", metavar="value", nargs="+",
@@ -572,21 +584,21 @@ def cli_argparser(kwargs, argparser=None):
                         help="pause before executing selected tests", type=str, required=False)
     parser.add_argument("--pause-after", dest="_pause_after", metavar="pattern", nargs="+",
                         help="pause after executing selected tests", type=str, required=False)
-    parser.add_argument("--random", dest="_random_order", action="store_true",
-                        help="randomize order of auto loaded tests", required=False)
+    parser.add_argument("--random", dest="_random", action="store_true",
+                        help="randomize order of auto loaded tests", required=False, default=None)
     parser.add_argument("--debug", dest="_debug", action="store_true",
-                        help="enable debugging mode", default=False)
+                        help="enable debugging mode", default=None)
     parser.add_argument("--no-colors", dest="_no_colors", action="store_true",
-                        help="disable terminal color highlighting", default=False)
+                        help="disable terminal color highlighting", default=None)
     parser.add_argument("--id", metavar="id", dest="_id", type=str, help="custom test id")
     parser.add_argument("-o", "--output", dest="_output", metavar="format", type=str,
-                        choices=["new-fails", "fails", "classic", "slick", "nice", "brisk", "quiet", "short", "manual", "dots", "raw"], default="nice",
+                        choices=["new-fails", "fails", "classic", "slick", "nice", "brisk", "quiet", "short", "manual", "dots", "raw"],
                         help="""stdout output format, choices are: ['new-fails','fails','classic','slick','nice','brisk','short','manual','dots','quiet','raw'],
                             default: 'nice'""")
     parser.add_argument("-l", "--log", dest="_log", metavar="file", type=str,
                         help="path to the log file where test output will be stored, default: uses temporary log file")
     parser.add_argument("--show-skipped", dest="_show_skipped", action="store_true",
-                        help="show skipped tests, default: False", default=False)
+                        help="show skipped tests, default: False", default=None)
     parser.add_argument("--repeat", dest="_repeat",
                         help=("number of times to repeat a test until it either fails or passes.\n"
                               "Where `pattern` is a test name pattern, "
@@ -602,10 +614,11 @@ def cli_argparser(kwargs, argparser=None):
                         choices=choices, nargs="+",
                         help=("rerun tests in the --reference log file.\n"
                               f"Where `result` is either {choices}"))
-    parser.add_argument("--individually", dest="_individually", action="store_true", default=False,
+    parser.add_argument("--individually", dest="_individually", action="store_true", default=None,
                         help="if --rerun is specified then rerun tests in the --reference log file individually.")
 
-    parser.add_argument("--parallel", dest="_parallel", type=onoff_type, default=True, choices=["yes", "no", "on", "off", 0, 1],
+    parser.add_argument("--parallel", dest="_parallel", type=onoff_type,
+        choices=["yes", "no", "on", "off", 0, 1],
         help="enable or disable parallelism for tests that support it, default: on")
     parser.add_argument("--parallel-pool", dest="_parallel_pool", metavar="size", type=count_type,
         help="for parallel tests force to use global parallel pool of the specified size")
@@ -614,9 +627,9 @@ def cli_argparser(kwargs, argparser=None):
         help="RSA private key PEM file that can be used to encrypt secrets.")
 
     exit_group = parser.add_mutually_exclusive_group()
-    exit_group.add_argument('--first-fail', dest="_first_fail", action='store_true',
+    exit_group.add_argument('--first-fail', dest="_first_fail", action='store_true', default=None,
         help="force all tests to be first fail and abort the run on the first failing test")
-    exit_group.add_argument('--test-to-end', dest="_test_to_end", action='store_true',
+    exit_group.add_argument('--test-to-end', dest="_test_to_end", action='store_true', default=None,
         help="force all tests to be test to end and continue the run even if one of the tests fails")
 
     if database_module:
@@ -635,23 +648,44 @@ def parse_cli_args(kwargs, parser):
     try:
         args, unknown = parser.parse_known_args()
         args = vars(args)
+        exc = None
+        default_config = os.path.join(os.path.expanduser("~"), ".testflows.yml")
+        configs = []
 
-        if args.get("_name"):
-            kwargs["name"] = args.pop("_name")
+        if os.path.exists(default_config):
+            configs.insert(0, open(default_config, "r"))
 
-        if args.get("_debug"):
-            settings.debug = True
-            args.pop("_debug")
+        if args.get("_config"):
+            configs += args.pop("_config")
 
+        try:
+            configs.reverse()
+            for config in configs:
+                obj = yaml.safe_load(config) or {}
+                _args = { f"_{k.replace('-','_')}" : v for k,v in obj.pop("test run", {}).items()}
+                _args.update(obj)
+                _args.update({k:v for k,v in args.items() if v is not None})
+                args = _args
+        except Exception as e:
+            exc = e
+        finally:
+            for config in configs:
+                config.close()
+
+        settings.debug = args.pop("_debug", None) or False
         debug_processed = True
 
-        if args.get("_no_colors"):
-            settings.no_colors = True
-            args.pop("_no_colors")
+        if exc is not None:
+            raise exc
 
         if unknown:
             raise ExitWithError(f"unknown argument {unknown}")
 
+        settings.no_colors = args.pop("_no_colors", None) or False
+
+        if args.get("_name"):
+            kwargs["name"] = args.pop("_name")
+
         if args.get("_id"):
             settings.test_id = args.get("_id")
             args.pop("_id")
@@ -667,18 +701,13 @@ def parse_cli_args(kwargs, parser):
         if os.path.exists(settings.write_logfile):
             os.remove(settings.write_logfile)
 
-        settings.output_format = args.pop("_output")
+        settings.output_format = args.pop("_output", None) or "nice"
 
         if args.get("_database"):
             settings.database = args.pop("_database")
 
-        if args.get("_show_skipped"):
-            settings.show_skipped = True
-            args.pop("_show_skipped")
-
-        if args.get("_random_order"):
-            settings.random_order = True
-            args.pop("_random_order")
+        settings.show_skipped = args.pop("_show_skipped", None) or False
+        settings.random_order = args.pop("_random", None) or False
 
         if args.get("_pause_before"):
             xflags = kwargs.get("xflags", {})
@@ -748,12 +777,14 @@ def parse_cli_args(kwargs, parser):
         if args.get("_private_key"):
             kwargs["private_key"] = args.pop("_private_key")
 
-        if args.get("_parallel") is False:
+        if args.get("_parallel") in ("no", 0, False):
             kwargs["flags"] = kwargs.get("flags", Flags())
             kwargs["flags"] |= NO_PARALLEL
 
         if args.get("_parallel_pool"):
-            settings.global_parallel_pool = ParallelPool(args.pop("_parallel_pool"))
+            pool_size = args.pop("_parallel_pool")
+            settings.global_thread_pool = GlobalThreadPoolExecutor(max_workers=pool_size)
+            settings.global_async_pool = GlobalAsyncPoolExecutor(max_workers=pool_size)
 
         if args.get("_repeat"):
             repeat = []
@@ -762,7 +793,7 @@ def parse_cli_args(kwargs, parser):
             kwargs["repeat"] = repeat
 
         if args.get("_rerun"):
-            rerun_individually = args.pop("_individually", False)
+            rerun_individually = args.pop("_individually", None) or False
             rerun = args.pop("_rerun")
 
             if not args.get("_reference"):
@@ -942,10 +973,37 @@ class TestDefinition(object):
         self.no_arguments = get(self.no_arguments, not args)
 
         def callable():
+            if is_running_in_event_loop():
+                raise RuntimeError("should not be running inside the async event loop")
+
             if test and isinstance(test, TestDecorator):
                 self.repeatable_func = test
                 with self as _test:
                     r = test(**self.kwargs["args"])
+                    if asyncio.iscoroutine(r):
+                        def _wrapper(r):
+                            return r
+                        with AsyncPoolExecutor() as executor:
+                            r = executor.submit(_wrapper, args=(r,)).result()
+                    if r is not None:
+                        value("return", value=r)
+                return _test
+            else:
+                with self as _test:
+                    pass
+                return _test
+
+        async def async_callable():
+            if test and isinstance(test, TestDecorator):
+                self.repeatable_func = test
+                with self as _test:
+                    if not asyncio.iscoroutinefunction(test.func):
+                        with ThreadPoolExecutor() as executor:
+                            def _wrapper(test):
+                                return test(**self.kwargs["args"])
+                            r = await asyncio.get_event_loop().run_in_executor(executor, _wrapper, (test,))
+                    else:
+                        r = await test(**self.kwargs["args"])
                     if r is not None:
                         value("return", value=r)
                 return _test
@@ -955,25 +1013,61 @@ class TestDefinition(object):
                 return _test
 
         current_test = current()
+        is_async = is_running_in_event_loop()
+        is_parallel = self.kwargs.get("flags", Flags()) & PARALLEL or self.parallel
 
         if current_test:
             if current_test.cflags & NO_PARALLEL:
                 pass
-            elif self.kwargs.get("flags", Flags()) & PARALLEL or self.parallel:
+
+            elif is_parallel:
                 self.kwargs["flags"] = self.kwargs.pop("flags", Flags()) | PARALLEL
                 executor = self.executor
-                if settings.global_parallel_pool is not None:
-                    executor = settings.global_parallel_pool
+                if is_async:
+                    if settings.global_async_pool is not None:
+                        executor = settings.global_async_pool
+                else:
+                    if settings.global_thread_pool is not None:
+                        executor = settings.global_thread_pool
+
                 if executor is None:
-                    executor = current_test.executor or ParallelPool()
+                    if is_async:
+                        executor = current_test.executor or AsyncPoolExecutor()
+                    else:
+                        executor = current_test.executor or ThreadPoolExecutor()
                     if current_test.executor is None:
                         current_test.executor = executor
+
                 if not executor.open:
                     executor.__enter__()
-                future = executor.submit(callable)
+
+                if isinstance(executor, AsyncPoolExecutor):
+                    if executor is settings.global_async_pool:
+                        future = executor.submit(async_callable, block=False)
+                    else:
+                        future = executor.submit(async_callable)
+                else:
+                    if executor is settings.global_thread_pool:
+                        future = executor.submit(callable, block=False)
+                    else:
+                        future = executor.submit(callable)
+
                 current_test.futures.append(future)
                 return future
 
+        if is_async:
+            if is_parallel:
+                future = asyncio.ensure_future(async_callable())
+                if current_test:
+                    current_test.futures.append(future)
+                return future
+            else:
+                return async_callable()
+        if is_parallel:
+            future = convert_result_to_concurrent_future(callable)
+            if current_test:
+                current_test.futures.append(future)
+            return future
         return callable()
 
     def __enter__(self):
@@ -1657,6 +1751,9 @@ class TestDecorator(object):
 
         self.func.kwargs = kwargs
 
+        if asyncio.iscoroutinefunction(self.func):
+            self.func.kwargs["flags"] = Flags(self.func.kwargs.pop("flags", None)) | ASYNC
+
         _type = self.type
         functools.update_wrapper(self, self.func)
         self.type = _type
@@ -1665,7 +1762,36 @@ class TestDecorator(object):
         if pargs:
             raise TypeError(f"only named arguments are allowed but {pargs} positional arguments were passed")
 
-        return self.__run__(**args)
+        if is_running_in_event_loop():
+            if not asyncio.iscoroutinefunction(self.func):
+                with ThreadPoolExecutor() as executor:
+                    def _wrapper():
+                        return self.__run__(**args)
+                    r = asyncio.get_event_loop().run_in_executor(executor, _wrapper)
+                    current().futures.append(r)
+                    return r
+
+        r = self.__run__(**args)
+        if asyncio.iscoroutine(r):
+            if not is_running_in_event_loop():
+                def _wrapper(r):
+                    return r
+                current_test = current()
+                executor = current_test.executor if current_test else None
+                if not isinstance(executor, AsyncPoolExecutor):
+                    with AsyncPoolExecutor() as executor:
+                        r = executor.submit(_wrapper, args=(r,)).result()
+                else:
+                    executor = settings.global_async_pool or executor
+
+                    if not executor.open:
+                        executor.__enter__()
+
+                    if executor is settings.global_async_pool:
+                        r = executor.submit(_wrapper, args=(r,), block=False).result()
+                    else:
+                        r = executor.submit(_wrapper, args=(r,)).result()
+        return r
 
     def __run__(self, **args):
         _check_parallel_context()
